// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
/***/
/*
It calculates the dihedral angle for each vertex of the mesh/ point in the point cloud
Outputs:
- DiAn: Dihedral angle array corresponding to each input vertex/point in the mesh.
Inputs:
- trimesh: 1-D array where every 3 values correspond to the vertex IDs of each triangular face of the mesh
- SZnodes: Number of input faces
- Nnodes: Number of input vertex/nodes/points
- MAXFaces: Max number of faces that a vertex can be part of. This is needed to keep memory allocation under control.
- xbuf: 1D buffer containing the x-coordinate of the point cloud/mesh vertices
- ybuf: 1D buffer containing the y-coordinate of the point cloud/mesh vertices
- zbuf: 1D buffer containing the z-coordinate of the point cloud/mesh vertices
- searchoffset: search radius in the 2-D grid to find potential points within pointspacing distance. (This is to avoid searching over millions of points everytime).
Version history: 2017-03-28	Marc Bosch	Initial version
*/
/***/
#include <math.h>
#include "mex.h"
#include <stdio.h>
#include <stdlib.h>
#include <limits>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{	
    if(nlhs > 1){
        mexErrMsgTxt("Function has one return value");
        return;
    }

    if(nrhs != 8){
        mexErrMsgTxt("Usage: [DiAn]=mexDihAngC(double *trimesh, double SZnodes, double Nnodes, double MAXFaces, double *xbuf, double *ybuf, double *zbuf, double searchoffset)");
        return;
    }
    const double eps = 0.01;
    /* get the image buffer*/    
    double *trimesh = (double*) mxGetPr(prhs[0]);	
    int* dims = (int*) mxGetDimensions(prhs[0]);
    int WorldDim = dims[0];//flipped since it comes transpose and row major
    int Nfaces = dims[1];

	double* pSZnodes = mxGetPr(prhs[1]);
 	int SZnodes = (int) *pSZnodes;

	double* pNnodes = mxGetPr(prhs[2]);
 	int Nnodes = (int) *pNnodes;
	
	double* pMAXFaces = mxGetPr(prhs[3]);
 	int MAXFaces = (int) *pMAXFaces;

	

	double *xbuf = (double*) mxGetPr(prhs[4]);
	double *ybuf = (double*) mxGetPr(prhs[5]);
	double *zbuf = (double*) mxGetPr(prhs[6]);
	
	double* psearchoffset = mxGetPr(prhs[7]);
 	int searchoffset = (int) *psearchoffset;
	
	//outputs
	
	plhs[0] = mxCreateDoubleMatrix(1, Nnodes, mxREAL);
    double *DiAn = mxGetPr(plhs[0]);  
    for (int i=0; i < Nnodes; i++)
        DiAn[i] = xbuf[i]; 

	
	//input mesh
	//temp variables and buffers




	int *c = (int*) malloc(SZnodes*sizeof(int));
	for (int i=0; i < SZnodes; i++)
		c[i] = 0;	
		
	//extend xbuf,... and xbufref,... to size=SZnodes/SZnodesref
	double * xbuf_prime= (double*) malloc(SZnodes*sizeof(double));
	double * ybuf_prime= (double*) malloc(SZnodes*sizeof(double));
	double * zbuf_prime= (double*) malloc(SZnodes*sizeof(double));
	int *pair = (int*) malloc(SZnodes*sizeof(int));
	int *flagunique = (int*) malloc(SZnodes*sizeof(int));
	int *unique_nodes = (int*) malloc(Nnodes*sizeof(int));
	for(int i=0;i<Nnodes;i++)
		unique_nodes[i]=0;
	for(int i=0;i<SZnodes;i++)
	{
		xbuf_prime[i]=0;ybuf_prime[i]=0;zbuf_prime[i]=0;
		pair[i]=-1;flagunique[i]=0;
	}
	for (int k=0;k<Nfaces;k++)//loop through all faces	
		for (int j=0;j<WorldDim;j++)
		{
			flagunique[(int)trimesh[k*WorldDim+j]-1] = 1;
		}
	int count=0;
	for(int i=0;i<SZnodes;i++)
		if(flagunique[i]==1)
			unique_nodes[count++] = i;
	
	for (int k=0;k<Nnodes;k++)//loop through all faces
	{
		int nodeid=unique_nodes[k];
		xbuf_prime[nodeid]=xbuf[k];
		ybuf_prime[nodeid]=ybuf[k];
		zbuf_prime[nodeid]=zbuf[k];
		pair[nodeid] = k;
	}

	//find for each node all the nodes connected by an edge



	double* DiAntemp = (double*) malloc(sizeof(double)*SZnodes);
	for(int i=0;i<SZnodes;i++)
		DiAntemp[i] = 0;
	double* facenormals = (double*) malloc(Nfaces*3*sizeof(double));
	for(int i=0;i<Nfaces*3;i++)
		facenormals[i] = 0;
	double vec1face[3];
	double vec2face[3];
	double v[3],v1[3],v2[3],v3[3];

	for (int k=0;k<Nfaces;k++)//loop through all faces
	{
		//store for each node all its neighbors (same face vertices)
		int nodeid=(int)trimesh[k*WorldDim]-1;
		int nodeid1=(int)trimesh[k*WorldDim+1]-1;
		int nodeid2=(int)trimesh[k*WorldDim+2]-1;
		vec1face[0] = xbuf_prime[nodeid1]-xbuf_prime[nodeid];
		vec1face[1] = ybuf_prime[nodeid1]-ybuf_prime[nodeid];
		vec1face[2] = zbuf_prime[nodeid1]-zbuf_prime[nodeid];
		vec2face[0] = xbuf_prime[nodeid2]-xbuf_prime[nodeid];
		vec2face[1] = ybuf_prime[nodeid2]-ybuf_prime[nodeid];
		vec2face[2] = zbuf_prime[nodeid2]-zbuf_prime[nodeid];
		//crossp
		v[0]=vec1face[1]*vec2face[2]-vec1face[2]*vec2face[1];
		v[1]=vec1face[0]*vec2face[2]-vec1face[2]*vec2face[0];
		v[2]=vec1face[0]*vec2face[1]-vec1face[1]*vec2face[0];
		double normv = sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
		normv=(normv==0)?1:normv;
		facenormals[k*3] = v[0]/normv;
		facenormals[k*3+1] = v[1]/normv;
		facenormals[k*3+2] = v[2]/normv;
	}

	//for each node... find all the edges it belongs to, and for each edge find both faces
	int klast=0;
	for (int k=0;k<Nfaces;k++)//loop through all faces
	{
		
		int nodeid=(int)trimesh[k*WorldDim]-1;			
		v1[0] = facenormals[3*k];
		v1[1] = facenormals[3*k+1];
		v1[2] = facenormals[3*k+2];
		for(int kk=1;kk<WorldDim;kk++)
		{
				
			int nodeidpair=(int)trimesh[k*WorldDim+kk]-1;	
			//find in the other faces the same edge
			int startface = MAX(k-searchoffset,0);
			int endface = MIN(k+searchoffset,Nfaces);
			count=1;
			int posface=0;
			for(int jj=startface;jj<endface;jj++)
			{
				if(jj!=k)
				{
					int nodeidcand=(int)trimesh[jj*WorldDim+0]-1;
					int nodeidcand1=(int)trimesh[jj*WorldDim+1]-1;
					int nodeidcand2=(int)trimesh[jj*WorldDim+2]-1;
					if((nodeidcand==nodeid && nodeidcand1==nodeidpair) || 
						(nodeidcand==nodeidpair && nodeidcand1==nodeid))//found edge
					{
						count++;
					}
					else if((nodeidcand==nodeid && nodeidcand2==nodeidpair) ||
						((nodeidcand==nodeidpair && nodeidcand2==nodeid)))//found edge
					{
							
						count++;
					}
					else if((nodeidcand1==nodeid && nodeidcand2==nodeidpair) ||
						((nodeidcand1==nodeidpair && nodeidcand2==nodeid)))//found edge
					{
							
						count++;
					}
					if (count==2)
					{
						posface=jj;
						break;
					}
				}
			}
			if (count>=2)//found both faces
			{
				v2[0] = facenormals[3*posface];
				v2[1] = facenormals[3*posface+1];
				v2[2] = facenormals[3*posface+2];
				//crossp
				v[0]=v1[1]*v2[2]-v1[2]*v2[1];
				v[1]=v1[0]*v2[2]-v1[2]*v2[0];
				v[2]=v1[0]*v2[1]-v1[1]*v2[0];
				//dotp
				//v3[0] = v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2];

				double theta = (180/3.1416)*(atan2(sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]),
					v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2]));
				/*if (theta < 0.0) 
				{
					theta += 360.0;
				}*/
				//to degrees
				DiAntemp[nodeid] += theta;
				DiAntemp[nodeidpair]+=theta;//DiAntemp[nodeid*MAXFaces+c[nodeid]];
				c[nodeidpair]++;
				c[nodeid]++;
				/*c[nodeid] = (c[nodeid]>=MAXFaces-1)?MAXFaces-1:c[nodeid];
				c[nodeidpair] = (c[nodeidpair]>=MAXFaces-1)?MAXFaces-1:c[nodeidpair];*/

			}
		}
	}
	//average over all edges of a node
	for(int i=0;i<SZnodes;i++)
		DiAn[pair[i]]=0;
	for(int i=0;i<SZnodes;i++)
	{
		/*for(int j=0;j<c[i];j++)
		{
			DiAn[pair[i]] += DiAntemp[i*MAXFaces+j];
		}*/
		if(c[i]==0)
			DiAn[pair[i]]=-1000;
		else
			DiAn[pair[i]]=DiAntemp[i]/(double)c[i];
	}



		
	free(xbuf_prime);
	free(ybuf_prime);
	free(zbuf_prime);	
    free(c);
	free(pair);
	free(flagunique);
	free(unique_nodes);

	


	
   
}
